@@ -22,6 +22,14 @@ file.
 
 Each configuration option has a default value.
 
+.. contents:: Table of Configuration Options
+    :local:
+
+A complete example TOML configuration:
+
+.. literalinclude:: ../config.toml.example
+
+
 Generic Options
 ===============
 
@@ -35,7 +43,17 @@ The AMQP broker to connect to. This URL should be in the format described by
 the :class:`pika.connection.URLParameters` documentation. This defaults to
 ``'amqp://?connection_attempts=3&retry_delay=5``.
 
-.. _conf-client-properties:
+.. note:: When using the Twisted consumer API, which the CLI does by default,
+          any connection-related setting won't apply as Twisted manages the
+          TCP/TLS connection.
+
+
+.. _conf-passive-declares:
+
+passive_declares
+----------------
+A boolean to specify if queues and exchanges should be declared passively (i.e
+checked, but not actually created on the server). Defaults to ``False``.
 
 
 .. _conf-tls:
@@ -60,6 +78,9 @@ when authenticating with the broker.
           provide these setting via the ``amqp_url`` setting using a URL-encoded
           JSON object. This setting is provided as a convenient way to avoid that.
 
+
+.. _conf-client-properties:
+
 client_properties
 -----------------
 A dictionary that describes the client to the AMQP broker. This makes it easy
@@ -74,9 +95,18 @@ arbitrary string keys and values. The default is::
     }
 
 Apps should set the ``app`` along with any additional keys they feel will help
-administrators when debugging application connections. Do not use the
-``product``, ``information``, and ``version`` keys as these will be set
-automatically.
+administrators when debugging application connections. At a minimum, the recommended
+fields are:
+
+* ``app_url``: The value of this key should be a URL to the upstream project for
+  the client.
+* ``app_contacts_email``: One or more emails of maintainers to contact with
+  questions (if, for example, a client is misbehaving, or a service disruption
+  is about to occur).
+
+Do not use the ``product``, ``information``, and ``version`` keys as these will
+be set automatically.
+
 
 .. _conf-exchanges:
 
@@ -117,6 +147,8 @@ log_config
 A dictionary describing the logging configuration to use, in a format accepted
 by :func:`logging.config.dictConfig`.
 
+.. note:: Logging is only configured for consumers, not for producers.
+
 
 Publisher Options
 =================
@@ -131,6 +163,14 @@ A string that identifies the exchange to publish to. The default is
 ``amq.topic``.
 
 
+.. _conf-topic-prefix:
+
+topic_prefix
+------------
+A string that will be prepended to topics on sent messages.
+This is useful to migrate from fedmsg, but should not be used otherwise.
+The default is an empty string.
+
 .. _sub-config:
 
 Consumer Options
@@ -146,16 +186,18 @@ A dictionary of queues that should be present in the broker. Each key should be
 a queue name, and the value should be a dictionary with the queue's configuration.
 Options are:
 
-* ``durable`` - whether or not the queue should survive a broker restart.
+* ``durable`` - whether or not the queue should survive a broker restart. This is
+  set to ``False`` for the default queue.
 
 * ``auto_delete`` - whether or not the queue should be deleted once the
-  consumer disconnects.
+  consumer disconnects. This is set to ``True`` for the default queue.
 
 * ``exclusive`` - whether or not the queue is exclusive to the current
-  connection.
+  connection. This is set to ``False`` for the default queue.
 
 * ``arguments`` - dictionary of arbitrary keyword arguments for the queue, which
-  depends on the broker in use and its extensions.
+  depends on the broker in use and its extensions. This is set to ``{}`` for the
+  default queue
 
 For example::
 
@@ -189,6 +231,8 @@ routing keys. For example::
 This would create two bindings for the ``my_queue`` queue, both to the
 ``amq.topic`` exchange. Consumers will consume from both queues.
 
+.. _conf-callback:
+
 callback
 --------
 The Python path of the callback. This should be in the format
@@ -205,7 +249,17 @@ consumer_config
 ---------------
 A dictionary for the consumer to use as configuration. The consumer should
 access this key in its callback for any configuration it needs. Defaults to
-an empty dictionary.
+an empty dictionary. If, for example, this dictionary contains the
+``print_messages`` key, the callback can access this configuration with::
+
+    from fedora_messaging import config
+
+    def callback(message):
+        if config.conf["consumer_config"]["print_messages"]:
+            print(message)
+
+
+.. _conf-qos:
 
 qos
 ---
@@ -225,13 +279,14 @@ meaning there is no limit. The default settings are::
 """
 from __future__ import unicode_literals
 
+import copy
 import logging
 import logging.config
 import os
 import uuid
 
 import pkg_resources
-import pytoml
+import toml
 
 from . import exceptions
 
@@ -244,7 +299,8 @@ _pika_version = pkg_resources.get_distribution("pika").version
 # A default, auto-deleted queue for consumers
 _default_queue_name = str(uuid.uuid4())
 
-#: A dictionary of application configuration defaults.
+#: The default configuration settings for fedora-messaging. This should not be
+#: modified and should be copied with :func:`copy.deepcopy`.
 DEFAULTS = dict(
     amqp_url="amqp://?connection_attempts=3&retry_delay=5",
     #: The default client properties reported to the AMQP broker in the "start-ok"
@@ -260,6 +316,8 @@ DEFAULTS = dict(
         ),
     },
     publish_exchange="amq.topic",
+    topic_prefix="",
+    passive_declares=False,
     exchanges={
         "amq.topic": {
             "type": "topic",
@@ -308,6 +366,91 @@ DEFAULTS = dict(
 )
 
 
+def validate_bindings(bindings):
+    """
+    Validate the bindings configuration.
+
+    Raises:
+        exceptions.ConfigurationException: If the configuration provided is of an
+            invalid format.
+    """
+    if not isinstance(bindings, (list, tuple)):
+        raise exceptions.ConfigurationException(
+            "bindings must be a list or tuple of dictionaries, but was a {}".format(
+                type(bindings)
+            )
+        )
+
+    for binding in bindings:
+        missing_keys = []
+        for key in ("queue", "exchange", "routing_keys"):
+            if key not in binding:
+                missing_keys.append(key)
+        if missing_keys:
+            raise exceptions.ConfigurationException(
+                "a binding is missing the following keys from its settings "
+                "value: {}".format(missing_keys)
+            )
+
+        if not isinstance(binding["routing_keys"], (list, tuple)):
+            raise exceptions.ConfigurationException(
+                "routing_keys must be a list or tuple, but was a {}".format(
+                    type(binding["routing_keys"])
+                )
+            )
+
+
+def validate_queues(queues):
+    """
+    Validate the queues configuration.
+
+    Raises:
+        exceptions.ConfigurationException: If the configuration provided is of an
+            invalid format.
+    """
+    if not isinstance(queues, dict):
+        raise exceptions.ConfigurationException(
+            "'queues' must be a dictionary mapping queue names to settings."
+        )
+
+    for queue, settings in queues.items():
+        if not isinstance(settings, dict):
+            raise exceptions.ConfigurationException(
+                "the {} queue in the 'queues' setting has a value of type {}, but it "
+                "should be a dictionary of settings.".format(queue, type(settings))
+            )
+        missing_keys = []
+        for key in ("durable", "auto_delete", "exclusive", "arguments"):
+            if key not in settings:
+                missing_keys.append(key)
+        if missing_keys:
+            raise exceptions.ConfigurationException(
+                "the {} queue is missing the following keys from its settings "
+                "value: {}".format(queue, missing_keys)
+            )
+
+
+def validate_client_properties(props):
+    """
+    Validate the client properties setting.
+
+    This will add the "version", "information", and "product" keys if they are
+    missing. All other keys are application-specific.
+
+    Raises:
+        exceptions.ConfigurationException: If any of the basic keys are overridden.
+    """
+    for key in ("version", "information", "product"):
+        # Nested dictionaries are not merged so key can be missing
+        if key not in props:
+            props[key] = DEFAULTS["client_properties"][key]
+        # Don't let users override these as they identify this library in AMQP
+        if props[key] != DEFAULTS["client_properties"][key]:
+            raise exceptions.ConfigurationException(
+                '"{}" is a reserved keyword in client_properties'.format(key)
+            )
+
+
 class LazyConfig(dict):
     """This class lazy-loads the configuration file."""
 
@@ -355,15 +498,9 @@ class LazyConfig(dict):
                     " {}".format(key, list(DEFAULTS.keys()))
                 )
 
-        for key in ("version", "information", "product"):
-            # Nested dictionaries are not merged so key can be missing
-            if key not in self["client_properties"]:
-                self["client_properties"][key] = DEFAULTS["client_properties"][key]
-            # Don't let users override these as they identify this library in AMQP
-            if self["client_properties"][key] != DEFAULTS["client_properties"][key]:
-                raise exceptions.ConfigurationException(
-                    '"{}" is a reserved keyword in client_properties'.format(key)
-                )
+        validate_queues(self["queues"])
+        validate_bindings(self["bindings"])
+        validate_client_properties(self["client_properties"])
 
     def load_config(self, config_path=None):
         """
@@ -375,7 +512,7 @@ class LazyConfig(dict):
         Otherwise, the path defaults to ``/etc/fedora-messaging/config.toml``.
         """
         self.loaded = True
-        config = DEFAULTS.copy()
+        config = copy.deepcopy(DEFAULTS)
 
         if config_path is None:
             if "FEDORA_MESSAGING_CONF" in os.environ:
@@ -387,12 +524,12 @@ class LazyConfig(dict):
             _log.info("Loading configuration from {}".format(config_path))
             with open(config_path) as fd:
                 try:
-                    file_config = pytoml.loads(fd.read())
+                    file_config = toml.load(fd)
                     for key in file_config:
                         config[key.lower()] = file_config[key]
-                except pytoml.core.TomlError as e:
-                    msg = "Failed to parse {}: error at line {}, column {}".format(
-                        config_path, e.line, e.col
+                except toml.TomlDecodeError as e:
+                    msg = "Failed to parse {}: error at line {}, column {}: {}".format(
+                        config_path, e.lineno, e.colno, e.msg
                     )
                     raise exceptions.ConfigurationException(msg)
         else:
@@ -403,5 +540,5 @@ class LazyConfig(dict):
         return self
 
 
-#: The application configuration dictionary.
+#: The configuration dictionary used by fedora-messaging and consumers.
 conf = LazyConfig()
