@@ -27,6 +27,7 @@ this class in a small Python package of its own.
 import datetime
 import json
 import logging
+import warnings
 import uuid
 
 import jsonschema
@@ -34,6 +35,7 @@ import pika
 import pkg_resources
 import pytz
 
+from . import config
 from .exceptions import ValidationError
 
 
@@ -221,34 +223,63 @@ class Message(object):
     from the user. This allows the schema to change and evolve without breaking
     the user-facing API.
 
-    A message class includes a topic. This topic is used by message consumers
-    to filter what messages they receive. Topics should be a string of words
-    separated by '.' characters, with a length limit of 255 bytes. Because of
-    this limit, it is best to avoid non-ASCII characters as well as
-    variable-length components where the total size of the topic would exceed
-    255 bytes.
-
-    Attributes:
-        id (six.text_type): The message id as a unicode string.
-        topic (six.text_type): The message topic as a unicode string.
-        headers_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
-            :func:`jsonschema.validate` to validate the message headers.
-        body_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
-            :func:`jsonschema.validate` to validate the message headers.
-        severity (int): An integer that indicates the severity of the message. This is
-            used to determine what messages to notify end users about and should be
-            :data:`DEBUG`, :data:`INFO`, :data:`WARNING`, or :data:`ERROR`.
-        queue (str): The name of the queue this message arrived through.
+    There are a number of properties that are intended to be overridden by
+    users.  These fields are used to sort messages for notifications or are
+    used to create human-readable versions of the messages. Properties that are
+    intended for this purpose are noted in their attribute documentation below.
 
     Args:
         headers (dict): A set of message headers. Consult the headers schema for
             expected keys and values.
         body (dict): The message body. Consult the body schema for expected keys
-            and values.
+            and values. This dictionary must be JSON-serializable by the default
+            serializer.
         topic (six.text_type): The message topic as a unicode string. If this is
-            not provided, the default topic for the class is used.
+            not provided, the default topic for the class is used. See the
+            attribute documentation below for details.
         properties (pika.BasicProperties): The AMQP properties. If this is not
-            provided, they will be generated.
+            provided, they will be generated. Most users should not need to provide
+            this, but it can be useful in testing scenarios.
+        severity (int): An integer that indicates the severity of the message. This is
+            used to determine what messages to notify end users about and should be
+            :data:`DEBUG`, :data:`INFO`, :data:`WARNING`, or :data:`ERROR`. The
+            default is :data:`INFO`, and can be set as a class attribute or on
+            an instance-by-instance basis.
+
+    Attributes:
+        id (six.text_type): The message id as a unicode string. This attribute is
+            automatically generated and set by the library and users should only
+            set it themselves in testing scenarios.
+        topic (six.text_type): The message topic as a unicode string. The topic
+            is used by message consumers to filter what messages they receive.
+            Topics should be a string of words separated by '.' characters,
+            with a length limit of 255 bytes. Because of this byte limit, it is
+            best to avoid non-ASCII character. Topics should start general and
+            get more specific each word. For example: "bodhi.update.kernel" is
+            a possible topic. "bodhi" identifies the application, "update"
+            identifies the message, and "kernel" identifies the package in the
+            update. This can be set at a class level or on a instance level.
+            Dynamic, specific topics that allow for fine-grain filtering are
+            preferred.
+        headers_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
+            :func:`jsonschema.validate` to validate the message headers. For
+            most users, the default definition should suffice.
+        body_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
+            :func:`jsonschema.validate` to validate the message body. The body_schema
+            is retrieved on a message instance so it is not required to be a
+            class attribute, although this is a convenient approach. Users are
+            also free to write the JSON schema as a file and load the file from
+            the filesystem or network if they prefer.
+        body (dict): The message body as a Python dictionary. This is validated by
+            the body schema before publishing and before consuming.
+        severity (int): An integer that indicates the severity of the message. This is
+            used to determine what messages to notify end users about and should be
+            :data:`DEBUG`, :data:`INFO`, :data:`WARNING`, or :data:`ERROR`. The
+            default is :data:`INFO`, and can be set as a class attribute or on
+            an instance-by-instance basis.
+        queue (str): The name of the queue this message arrived through. This
+            attribute is set automatically by the library and users should never
+            set it themselves.
     """
 
     severity = INFO
@@ -275,9 +306,12 @@ class Message(object):
     def __init__(
         self, body=None, headers=None, topic=None, properties=None, severity=None
     ):
-        self._body = body or {}
+        self.body = body or {}
+
         if topic:
+            # Default is "" on base class
             self.topic = topic
+
         headers = headers or {}
         if severity:
             self.severity = severity
@@ -346,36 +380,55 @@ class Message(object):
     @property
     def _encoded_routing_key(self):
         """The encoded routing key used to publish the message on the broker."""
-        return self.topic.encode("utf-8")
+        topic = self.topic
+        if config.conf["topic_prefix"]:
+            topic = ".".join((config.conf["topic_prefix"].rstrip("."), topic))
+        return topic.encode("utf-8")
 
     @property
     def _encoded_body(self):
         """The encoded body used to publish the message."""
-        return json.dumps(self._body).encode("utf-8")
+        return json.dumps(self.body).encode("utf-8")
 
     def __repr__(self):
         """
         Provide a printable representation of the object that can be passed to func:`eval`.
         """
         return "{}(id={}, topic={}, body={})".format(
-            self.__class__.__name__, repr(self.id), repr(self.topic), repr(self._body)
+            self.__class__.__name__, repr(self.id), repr(self.topic), repr(self.body)
         )
 
     def __eq__(self, other):
         """
         Two messages of the same class with the same topic, headers, and body are equal.
 
+        The "sent-at" header is excluded from the equality check as this is set
+        automatically and is dependent on when the object is created.
+
         Args:
             other (object): The object to check for equality.
 
         Returns:
             bool: True if the messages are equal.
         """
+        if not isinstance(other, self.__class__):
+            return False
+
+        headers = self._headers.copy()
+        other_headers = other._headers.copy()
+        try:
+            del headers["sent-at"]
+        except KeyError:
+            pass
+        try:
+            del other_headers["sent-at"]
+        except KeyError:
+            pass
+
         return (
-            isinstance(other, self.__class__)
-            and self.topic == other.topic
-            and self._body == other._body
-            and self._headers == other._headers
+            self.topic == other.topic
+            and self.body == other.body
+            and headers == other_headers
         )
 
     def validate(self):
@@ -403,9 +456,9 @@ class Message(object):
             jsonschema.validate(self._headers, schema)
         for schema in (self.body_schema, Message.body_schema):
             _log.debug(
-                'Validating message body "%r" with schema "%r"', self._body, schema
+                'Validating message body "%r" with schema "%r"', self.body, schema
             )
-            jsonschema.validate(self._body, schema)
+            jsonschema.validate(self.body, schema)
 
     @property
     def summary(self):
@@ -415,6 +468,10 @@ class Message(object):
         This should provide a short summary of the message, much like the subject line
         of an email.
 
+        .. note:: Sub-classes should override this method. It is used to create
+            the subject of email notifications, IRC notification, and by other
+            tools to display messages to humans in short form.
+
         The default implementation is to simply return the message topic.
         """
         return self.topic
@@ -423,11 +480,13 @@ class Message(object):
         """
         A human-readable representation of this message.
 
-        This should provide a detailed representation of the message, much like the body
-        of an email.
+        This should provide a detailed, long-form representation of the
+        message. The default implementation is to format the raw message id,
+        topic, headers, and body.
 
-        The default implementation is to format the raw message id, topic, headers, and
-        body. Applications use this to present messages to users.
+        .. note:: Sub-classes should override this method. It is used to create
+            the body of email notifications and by other tools to display messages
+            to humans.
         """
         return "Id: {i}\nTopic: {t}\nHeaders: {h}\nBody: {b}".format(
             i=self.id,
@@ -435,7 +494,7 @@ class Message(object):
             h=json.dumps(
                 self._headers, sort_keys=True, indent=4, separators=(",", ": ")
             ),
-            b=json.dumps(self._body, sort_keys=True, indent=4, separators=(",", ": ")),
+            b=json.dumps(self.body, sort_keys=True, indent=4, separators=(",", ": ")),
         )
 
     @property
@@ -443,6 +502,9 @@ class Message(object):
         """
         An URL to the action that caused this message to be emitted.
 
+        .. note:: Sub-classes should override this method if there is a URL
+            associated with message.
+
         Returns:
             str or None: A relevant URL.
         """
@@ -452,6 +514,10 @@ class Message(object):
     def app_icon(self):
         """An URL to the icon of the application that generated the message.
 
+        .. note:: Sub-classes should override this method if their application
+            has an icon and they wish that image to appear in applications that
+            consume messages.
+
         Returns:
             str or None: The URL to the app's icon.
         """
@@ -461,6 +527,9 @@ class Message(object):
     def agent_avatar(self):
         """An URL to the avatar of the user who caused the action.
 
+        .. note:: Sub-classes should override this method if the message was
+            triggered by a particular user.
+
         Returns:
             str or None: The URL to the user's avatar.
         """
@@ -470,6 +539,10 @@ class Message(object):
     def usernames(self):
         """List of users affected by the action that generated this message.
 
+        .. note:: Sub-classes should override this method if the message pertains
+            to a user or users. The data returned from this property is used to
+            filter notifications.
+
         Returns:
             list(str): A list of affected usernames.
         """
@@ -477,7 +550,11 @@ class Message(object):
 
     @property
     def packages(self):
-        """List of packages affected by the action that generated this message.
+        """List of RPM packages affected by the action that generated this message.
+
+        .. note:: Sub-classes should override this method if the message pertains
+            to one or more RPM packages. The data returned from this property
+            is used to filter notifications.
 
         Returns:
             list(str): A list of affected package names.
@@ -488,6 +565,10 @@ class Message(object):
     def containers(self):
         """List of containers affected by the action that generated this message.
 
+        .. note:: Sub-classes should override this method if the message pertains
+            to one or more container images. The data returned from this property
+            is used to filter notifications.
+
         Returns:
             list(str): A list of affected container names.
         """
@@ -497,6 +578,10 @@ class Message(object):
     def modules(self):
         """List of modules affected by the action that generated this message.
 
+        .. note:: Sub-classes should override this method if the message pertains
+            to one or more modules. The data returned from this property is
+            used to filter notifications.
+
         Returns:
             list(str): A list of affected module names.
         """
@@ -506,7 +591,131 @@ class Message(object):
     def flatpaks(self):
         """List of flatpaks affected by the action that generated this message.
 
+        .. note:: Sub-classes should override this method if the message pertains
+            to one or more flatpaks. The data returned from this property is
+            used to filter notifications.
+
         Returns:
             list(str): A list of affected flatpaks names.
         """
         return []
+
+    @property
+    def _body(self):
+        warnings.warn(
+            "The '_body' property has been renamed to 'body'.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.body
+
+    @_body.setter
+    def _body(self, value):
+        warnings.warn(
+            "The '_body' property has been renamed to 'body'.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        self.body = value
+
+
+#: The schema for each JSON object produced by :func:`dumps`, consumed by
+#: :func:`loads`, and expected by CLI commands like "fedora-messaging publish".
+SERIALIZED_MESSAGE_SCHEMA = {
+    "$schema": "http://json-schema.org/draft-04/schema#",
+    "description": "Schema for the JSON object used to represent messages in a file",
+    "type": "object",
+    "properties": {
+        "topic": {"type": "string", "description": "The message topic"},
+        "headers": {
+            "type": "object",
+            "properties": Message.headers_schema["properties"],
+            "description": "The message headers",
+        },
+        "id": {"type": "string", "description": "The message's UUID."},
+        "body": {"type": "object", "description": "The message body."},
+        "queue": {
+            "type": "string",
+            "description": "The queue the message arrived on, if any.",
+        },
+    },
+    "required": ["topic", "headers", "id", "body", "queue"],
+}
+
+
+def dumps(messages):
+    """
+    Serialize messages to a file format acceptable for :func:`loads` or for the
+    publish CLI command. The format is a string where each line is a JSON
+    object that conforms to the :data:`SERIALIZED_MESSAGE_SCHEMA` format.
+
+    Args:
+        messages (list or Message): The messages to serialize. Each message in
+            the messages is subclass of Message.
+
+    Returns:
+        str: Serialized messages.
+
+    Raises:
+        ValidationError: If one of the messages provided doesn't conform to its
+            schema.
+    """
+    if isinstance(messages, Message):
+        messages = [messages]
+
+    serialized_messages = []
+    for message in messages:
+        try:
+            message.validate()
+        except (jsonschema.exceptions.ValidationError, AttributeError) as e:
+            raise ValidationError(e)
+        m = {
+            "topic": message.topic,
+            "headers": message._headers,
+            "id": message.id,
+            "body": message.body,
+            "queue": message.queue,
+        }
+        serialized_messages.append(json.dumps(m, ensure_ascii=False, sort_keys=True))
+
+    return "\n".join(serialized_messages) + "\n"
+
+
+def loads(serialized_messages):
+    """
+    Deserialize messages from a file format produced by :func:`dumps`.  The
+    format is a string where each line is a JSON object that conforms to the
+    :data:`SERIALIZED_MESSAGE_SCHEMA` format.
+
+    Args:
+        serialized_messages (str): A string made up of a JSON object per line.
+
+    Returns:
+        list: Deserialized message objects.
+
+    Raises:
+        ValidationError: If the string isn't formatted properly or message
+            doesn't pass the message schema validation
+    """
+    messages = []
+    for serialized_message in serialized_messages.splitlines():
+        try:
+            message_dict = json.loads(serialized_message)
+        except ValueError as e:
+            raise ValidationError(e)
+        try:
+            jsonschema.validate(message_dict, SERIALIZED_MESSAGE_SCHEMA)
+        except jsonschema.exceptions.ValidationError as e:
+            raise ValidationError(e)
+        MessageClass = get_class(message_dict["headers"]["fedora_messaging_schema"])
+        message = MessageClass(
+            body=message_dict["body"],
+            topic=message_dict["topic"],
+            headers=message_dict["headers"],
+            severity=message_dict["headers"]["fedora_messaging_severity"],
+        )
+        message.queue = message_dict["queue"]
+        message.id = message_dict["id"]
+        messages.append(message)
+
+    return messages
